/*
 * Copyright (c) 2016-2017, Inversoft Inc., All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific
 * language governing permissions and limitations under the License.
 */

'use strict';

const RESTClient = require('./RESTClient.js');
require('promise');

const PassportClient = function(apiKey, host) {
  this.apiKey = apiKey;
  this.host = host;
};

PassportClient.constructor = PassportClient;
//noinspection JSUnusedGlobalSymbols
PassportClient.prototype = {

  /**
   * Takes an action on a user. The user being actioned is called the "actionee" and the user taking the action is
   * called the "actioner". Both user ids are required. You pass the actionee's user id into the method and the
   * actioner's is put into the request object.
   *
   * @param {String} actioneeUserId The actionee's user id.
   * @param {Object} request        The action request that includes all of the information about the action being taken
   *                       including the id of the action, any options and the duration (if applicable).
   * @return {Promise} A Promise for the Passport call.
   */
  actionUser: function(actioneeUserId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/action')
          .urlSegment(actioneeUserId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Cancels the user action.
   *
   * @param {string} actionId The action id of the action to cancel.
   * @param {Object} request  The action request that contains the information about the cancellation.
   * @return {Promise} A Promise for the Passport call.
   */
  cancelAction: function(actionId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('api/user/action')
          .urlSegment(actionId)
          .setJSONBody(request)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Changes a user's password using the verification id. This usually occurs after an email has been sent to the user
   * and they clicked on a link to reset their password.
   *
   * @param {string} verificationId The uuid sent to a user in an email.
   * @param {Object} request The change password request.
   * @return {Promise} A Promise for the Passport call.
   */
  changePassword: function(verificationId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/change-password')
          .urlSegment(verificationId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Adds a comment to the user's account.
   *
   * @param {Object} request The comment request that contains all of the information used to add the comment to the
   *     user.
   * @return {Promise} A Promise for the Passport call.
   */
  commentOnUser: function(request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/comment')
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates the given application.
   *
   * @param {?string} id The id of the new application, or null to allow Passport to generate a unique Id.
   * @param {Object} request The application creation request.
   * @return {Promise} A Promise for the Passport call.
   */
  createApplication: function(id, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/application')
          .urlSegment(id)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates a new role for an application. You must specify the id of the application you are creating the role for.
   * You can optionally specify an id for the role inside the ApplicationRole object itself, but this is not required.
   *
   * @param {string} applicationId The id of the application to create the role on.
   * @param {Object} request       The application request that contains all of the information used to create the role.
   * @return {Promise} A Promise for the Passport call.
   */
  createApplicationRole: function(applicationId, request) {
    this._requireNonNull(applicationId, 'applicationId');
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlSegment('role')
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates an audit log with the message and user name (usually an email). Audit logs should be written anytime you
   * make changes to the Passport database. When using the Passport Backend web interface, any changes are automatically
   * written to the audit log. However, if you are accessing the API, you must write the audit logs yourself.
   *
   * @param {string} message    The message for the audit log.
   * @param {string} insertUser The user that took the action being logged. Generally an email or username.
   * @return {Promise} A Promise for the Passport call.
   */
  createAuditLog: function(message, insertUser) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/system/audit-log')
          .urlSegment('role')
          .setJSONBody({'auditLog': {'insertUser': insertUser, 'message': message}})
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates an email template. You can optionally specify an id for the email template when calling this method, but
   * it
   * is not required.
   *
   * @param {?string} emailTemplateId The id for the template, or null to allow Passport to generate a unique Id.
   * @param {Object} request  The email template request that contains all of the information used to create the email
   *     template.
   * @return {Promise} A Promise for the Passport call.
   */
  createEmailTemplate: function(emailTemplateId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/email/template')
          .urlSegment(emailTemplateId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates a user with an optional id.
   *
   * @param {?string} userId The id for the user, or null to allow Passport to generate a unique Id.
   * @param {Object} request The user request that contains all of the information used to create the user.
   * @return {Promise} A Promise for the Passport call.
   */
  createUser: function(userId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates a user action. This action cannot be taken on a user until this call successfully returns. Anytime after
   * that the user action can be applied to any user.
   *
   * @param {?string} userActionId The id for the user action, or null to allow Passport to generate a unique Id.
   * @param {Object} request      The user action request that contains all of the information used to create the user
   *                              action.
   * @return {Promise} A Promise for the Passport call.
   */
  createUserAction: function(userActionId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates a user reason. This user action reason cannot be used when actioning a user until this call completes
   * successfully. Anytime after that the user action reason can be used.
   *
   * @param {?string} userActionReasonId The id for the user action reason, or null to allow Passport to generate a
   *     unique Id.
   * @param {Object} request The user action reason request that contains all of the information used to create the
   *     user action reason.
   * @return {Promise} A Promise for the Passport call.
   */
  createUserActionReason: function(userActionReasonId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action-reason')
          .urlSegment(userActionReasonId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates a webhook. You can optionally specify an id for the webhook when calling this
   * method, but it is not required.
   *
   * @param {?string} webhookId The id for the webhook, or null to allow Passport to generate a unique Id.
   * @param {Object} request   The webhook request that contains all of the information used to create the webhook.
   * @return {Promise} A Promise for the Passport call.
   */
  createWebhook: function(webhookId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/webhook')
          .urlSegment(webhookId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deactivates the application with the given id.
   *
   * @param {string} applicationId The id of the application to deactivate.
   * @return {Promise} A Promise for the Passport call.
   */
  deactivateApplication: function(applicationId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deactivates the user with the given id.
   *
   * @param {string} userId The id of the user to deactivate.
   * @return {Promise} A Promise for the Passport call.
   */
  deactivateUser: function(userId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deactivates the user action with the given id.
   *
   * @param {string} userActionId The id of the user action to deactivate.
   * @return {Promise} A Promise for the Passport call.
   */
  deactivateUserAction: function(userActionId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Hard deletes an application. This is a dangerous operation and should not be used in most circumstances. This will
   * delete the application, any registrations for that application, metrics and reports for the application, all the
   * roles for the application, and any other data associated with the application. This operation could take a very
   * long time, depending on the amount of data in your database.
   *
   * @param {string} applicationId The id of the application to be deleted
   * @return {Promise} A Promise for the Passport call.
   */
  deleteApplication: function(applicationId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri("/api/application")
          .urlSegment(applicationId)
          .urlParameter("hardDelete", true)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Hard deletes an application role. This is a dangerous operation and should not be used in most circumstances. This
   * permanently removes the given role from all users that had it.
   *
   * @param {string} applicationId The id of the application that contains the role.
   * @param {string} roleId        The id of the role to delete.
   * @return {Promise} A Promise for the Passport call.
   */
  deleteApplicationRole: function(applicationId, roleId) {
    this._requireNonNull(applicationId, 'applicationId');
    return new Promise((resolve, reject) => {
      this._start()
          .uri("/api/application")
          .urlSegment(applicationId)
          .urlSegment('role')
          .urlSegment(roleId)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deletes the email template for the given id.
   *
   * @param {string} emailTemplateId The id of the email template to delete.
   * @return {Promise} A Promise for the Passport call.
   */
  deleteEmailTemplate: function(emailTemplateId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri("/api/email/template")
          .urlSegment(emailTemplateId)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deletes the user registration for the given user and application.
   *
   * @param {string} userId        The id of the user whose registration is being deleted.
   * @param {string} applicationId The id of the application to remove the registration for.
   * @return {Promise} A Promise for the Passport call.
   */
  deleteRegistration: function(userId, applicationId) {
    this._requireNonNull(userId, 'userId');
    return new Promise((resolve, reject) => {
      this._start()
          .uri("/api/user/registration")
          .urlSegment(userId)
          .urlSegment(applicationId)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deletes the user for the given id. This permanently deletes all information, metrics, reports and data associated
   * with the user.
   *
   * @param {string} userId The id of the user to be deleted
   * @return {Promise} A Promise for the Passport call.
   */
  deleteUser: function(userId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri("/api/user")
          .urlSegment(userId)
          .urlParameter("hardDelete", true)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deletes the user action for the given id. This permanently deletes the user action and also any history and logs of
   * the action being applied to any users.
   *
   * @param {string} userActionId The id of the user action to delete.
   * @return {Promise} A Promise for the Passport call.
   */
  deleteUserAction: function(userActionId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri("/api/user-action")
          .urlSegment(userActionId)
          .urlParameter("hardDelete", true)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deletes the user action reason for the given id.
   *
   * @param {string} userActionReasonId The id of the user action reason to delete.
   * @return {Promise} A Promise for the Passport call.
   */
  deleteUserActionReason: function(userActionReasonId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri("/api/user-action-reason")
          .urlSegment(userActionReasonId)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Deletes the webhook for the given id.
   *
   * @param webhookId The id of the webhook to delete.
   * @return {Promise} A Promise for the Passport call.
   */
  deleteWebhook: function(webhookId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri("/api/webhook")
          .urlSegment(webhookId)
          .delete()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Exchange a refresh token for a new Access Token (JWT).
   *
   * @param request The refresh request.
   * @return {Promise} A Promise for the Passport call.
   */
  exchangeRefreshTokenForAccessToken: function(request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/jwt/refresh')
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Begins the forgot password sequence, which kicks off an email to the user so that they can reset their password.
   *
   * @param {Object} request The forgot password request.
   * @return {Promise} A Promise for the Passport call.
   */
  forgotPassword: function(request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/forgot-password')
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Bulk imports multiple users. This does some validation, but then tries to run batch inserts of users. This reduces
   * latency when inserting lots of users. Therefore, the error response might contain some information about failures,
   * but it will likely be pretty generic.
   *
   * @param {Object} request The request that contains all of the information about all of the users to import.
   * @return {Promise} A Promise for the Passport call.
   */
  importUsers: function(request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/import')
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Logs a user in.
   *
   * @param {Object} request contains the user credentials to log in.
   * @return {Promise} A Promise for the Passport call.
   */
  login: function(request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/login')
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Sends a ping to Passport indicating that the user was automatically logged into an application. When using
   * Passport's SSO or your own, you should call this if the user is already logged in centrally, but accesses an
   * application where they no longer have a session. This helps correctly track login counts, times and helps with
   * reporting.
   *
   * @param {string} userId          The id of the user that was logged in.
   * @param {string} applicationId   The id of the application that they logged into.
   * @param {?string} [callerIPAddress] The optional IP address of the end-user that is logging in. If a null value is
   *                  provided the IP address will be that of the client or last proxy that sent the request.
   * @return {Promise} A Promise for the Passport call.
   */
  loginPing: function(userId, applicationId, callerIPAddress) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/login')
          .urlSegment(userId)
          .urlSegment(applicationId)
          .urlParameter('ipAddress', callerIPAddress)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Modifies a temporal user action by changing the expiration of the action and optionally adding a comment to the
   * action.
   *
   * @param {string} actionId The id of the action to modify. This is technically the user action log id.
   * @param {Object} request  The request that contains all of the information about the modification.
   * @return {Promise} A Promise for the Passport call.
   */
  modifyAction: function(actionId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/action')
          .urlSegment(actionId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Reactivates the application for the given id.
   *
   * @param {string} applicationId The id of the application to reactivate.
   * @return {Promise} A Promise for the Passport call.
   */
  reactivateApplication: function(applicationId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlParameter('reactivate', true)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Reactivates the user action for the given id.
   *
   * @param {string} userId The id of the user action to reactivate.
   * @return {Promise} A Promise for the Passport call.
   */
  reactivateUser: function(userId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .urlParameter('reactivate', true)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Reactivates the user action for the given id.
   *
   * @param {string} userActionId The id of the user action to reactivate.
   * @return {Promise} A Promise for the Passport call.
   */
  reactivateUserAction: function(userActionId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .urlParameter('reactivate', true)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Registers a user for an application. If you provide the User and the UserRegistration object on this request, it
   * will create the user as well as register them for the application. This is called a Full Registration. However, if
   * you only provide the UserRegistration object, then the user must already exist and they will be registered for the
   * application. The user id can also be provided and it will either be used to look up an existing user or it will be
   * used for the newly created User.
   *
   * @param {?string} userId The id of the user being registered for the application. If performing a Full Registration
   *                  (creating a new user and registration), this parameter is not required and if left null a user Id
   *                  will be generated. If you omit the User object from the request and are registering an existing
   *                  user, this parameter is required.
   * @param {Object} request contains the user information to register a user with Passport.
   * @return {Promise} A Promise for the Passport call.
   */
  register: function(userId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/registration')
          .urlSegment(userId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Re-sends the verification email to the user.
   *
   * @param {string} email The email address of the user that needs a new verification email.
   * @return {Promise} A Promise for the Passport call.
   */
  resendEmailVerification: function(email) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/verify-email')
          .urlParameter('email', email)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves a single action for the given id.
   *
   * @param {string} actionId The id of the action to retrieve.
   * @return {Promise} A Promise for the Passport call.
   */
  retrieveAction: function(actionId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/action')
          .urlSegment(actionId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves all of the actions for the user with the given id.
   *
   * @param {string} userId The id of the user to fetch the actions for.
   * @return {Promise} A Promise for the Passport call.
   */
  retrieveActions: function(userId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/action')
          .urlParameter('userId', userId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieve all of the applications
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveApplications: function() {
    return this.retrieveApplication();
  },

  /**
   * Retrieves the application with the given id or all of the applications if the Id is null.
   *
   * @param {?string} [applicationId] The UUID of the application to retrieve, if omitted or set to null all applications are returned.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveApplication: function(applicationId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the daily active user report between the two instants. If you specify an application id, it will only
   * return the daily active counts for that application.
   *
   * @param {?string} [applicationId] The application id, if set to null all applications will be included in the result.
   * @param {number} start         The start instant as UTC milliseconds since Epoch.
   * @param {number} end           The end instant as UTC milliseconds since Epoch.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveDailyActiveReport: function(applicationId, start, end) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/report/daily-active-user')
          .urlParameter("start", start)
          .urlParameter("end", end)
          .urlParameter("applicationId", applicationId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the email template for the given id. If you don't specify the id, this will return all of the email
   * templates.
   *
   * @param {?string} [emailTemplateId] The id of the email template. If omitted or set to null all email templates are returned.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveEmailTemplate: function(emailTemplateId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/email/template')
          .urlSegment(emailTemplateId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Creates a preview of the email template provided in the request. This allows you to preview an email template that
   * hasn't been saved to the database yet. The entire email template does not need to be provided on the request. This
   * will create the preview based on whatever is given.
   *
   * @param {Object} request The request that contains the email template and optionally a locale to render it in.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveEmailTemplatePreview: function(request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/email/template/preview')
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves all of the email templates.
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveEmailTemplates: function() {
    return this.retrieveEmailTemplate();
  },

  /**
   * Retrieves all of the applications that are currently inactive.
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveInactiveApplications: function() {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/application')
          .urlParameter('inactive', true)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves all of the user actions that are currently inactive.
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveInactiveUserActions: function() {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action')
          .urlParameter('inactive', true)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the Public Key configured for verifying JSON Web Tokens (JWT) by the key Id. If the key Id is provided a
   * single public key will be returned if one is found by that id. If the optional parameter key Id is not provided all
   * public keys will be returned.
   *
   * @param {?string} [keyId] The id of the public key. If omitted or set to null all public keys will be returned.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveJwtPublicKey: function(keyId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/jwt/public-key')
          .urlSegment(keyId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves all Public Keys configured for verifying JSON Web Tokens (JWT).
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveJwtPublicKeys: function() {
    return this.retrieveJwtPublicKey();
  },

  /**
   * Retrieves the login report between the two instants. If you specify an application id, it will only return the
   * login counts for that application.
   *
   * @param {?string} applicationId The application id, if set to null all applications will be used to build the report.
   * @param {number} start         The start instant as UTC milliseconds since Epoch.
   * @param {number} end           The end instant as UTC milliseconds since Epoch.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveLoginReport: function(applicationId, start, end) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/report/login')
          .urlParameter("start", start)
          .urlParameter("end", end)
          .urlParameter("applicationId", applicationId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the monthly active user report between the two instants. If you specify an application id, it will only
   * return the monthly active counts for that application.
   *
   * @param {?string} applicationId The application id, if set to null all applications will be used to build the report.
   * @param {number} start         The start instant as UTC milliseconds since Epoch.
   * @param {number} end           The end instant as UTC milliseconds since Epoch.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveMonthlyActiveReport: function(applicationId, start, end) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/report/monthly-active-user')
          .urlParameter("start", start)
          .urlParameter("end", end)
          .urlParameter("applicationId", applicationId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the refresh tokens that belong to the user with the given id.
   *
   * @param {string} userId The id of the user.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveRefreshTokens: function(userId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/jwt/refresh')
          .urlParameter('userId', userId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the user registration for the user with the given id and the given application id.
   *
   * @param {string} userId        The id of the user.
   * @param {string} applicationId The id of the application.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveRegistration: function(userId, applicationId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/registration')
          .urlSegment(userId)
          .urlSegment(applicationId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the registration report between the two instants. If you specify an application id, it will only return
   * the login counts for that application.
   *
   * @param {?string} applicationId The application id, if set to null all applications will be used to build the report.
   * @param {number} start         The start instant as UTC milliseconds since Epoch.
   * @param {number} end           The end instant as UTC milliseconds since Epoch.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveRegistrationReport: function(applicationId, start, end) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/report/registration')
          .urlParameter("start", start)
          .urlParameter("end", end)
          .urlParameter("applicationId", applicationId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the SystemConfiguration.
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveSystemConfiguration: function() {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/system-configuration')
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the totals report. This contains all of the total counts for each application and the global registration
   * count.
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveTotalReport: function() {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/report/totals')
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the user for the given id.
   *
   * @param {string} userId The id of the user.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUser: function(userId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the user action for the given id. If you pass in null for the id, this will return all of the user
   * actions.
   *
   * @param {?string} [userActionId] The id of the user action, if omitted or set to null all of the user actions will be returned.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUserAction: function(userActionId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves all User Actions.
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUserActions: function() {
    return this.retrieveUserAction();
  },

  /**
   * Retrieves the user action reason for the given id. If you pass in null for the id, this will return all of the user
   * action reasons.
   *
   * @param {?string} [userActionReasonId] The id of the user action, if omitted or set to null all user action reasons are returned.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUserActionReason: function(userActionReasonId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action-reason')
          .urlSegment(userActionReasonId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves all of the User Action Reasons.
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUserActionReasons: function() {
    return this.retrieveUserActionReason();
  },

  /**
   * Retrieves the user for the given email.
   *
   * @param {string} email The email of the user.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUserByEmail: function(email) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user')
          .urlParameter('email', email)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the user for the given username.
   *
   * @param {string} username The username of the user.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUserByUsername: function(username) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user')
          .urlParameter('username', username)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves all of the comments for the user with the given id.
   *
   * @param {string} userId The id of the user.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUserComments: function(userId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/comment')
          .urlSegment(userId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the the last number of login records for a user.
   *
   * @param userId The user's id.
   * @param offset The initial record. e.g. 0 is the last login, 100 will be the 100th most recent login.
   * @param {?number} [limit=10] The number of records to retrieve, optional and the default value is 10.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveUserLoginReport: function(userId, offset, limit) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/report/user-login')
          .urlParameter('userId', userId)
          .urlParameter('offset', offset)
          .urlParameter('limit', limit)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the webhook for the given id. If you pass in null for the id, this will return all the
   * webhooks.
   *
   * @param {?string} [webhookId] The id of the webhook, if omitted or set to null all webhooks will be returned.
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveWebhook: function(webhookId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/webhook')
          .urlSegment(webhookId)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves all of the webhooks
   *
   * @returns {Promise} A Promise for the Passport call.
   */
  retrieveWebhooks: function() {
    return this.retrieveWebhook();
  },

  /**
   * Searches the audit logs with the specified criteria and pagination.
   *
   * @param search The search criteria and pagination information.
   * @returns {Promise} A Promise for the Passport call.
   */
  searchAuditLogs: function(search) {
    return new Promise((resolve, reject) => {
      //noinspection JSUnresolvedVariable
      this._start()
          .uri('/api/system/audit-log')
          .urlParameter('search.user', search.user)
          .urlParameter('search.message', search.message)
          .urlParameter('search.end', search.end)
          .urlParameter('search.start', search.start)
          .urlParameter('search.orderBy', search.orderBy)
          .urlParameter('search.startRow', search.startRow)
          .urlParameter('search.numberOfResults', search.numberOfResults)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Searches the audit logs with the specified criteria and pagination.
   *
   * @param {string[]} ids The user ids to search for.
   * @returns {Promise} A Promise for the Passport call.
   */
  searchUsers: function(ids) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/search')
          .urlParameter('ids', ids)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Retrieves the users for the given search criteria and pagination.
   *
   * @param {Object} search The search criteria and pagination constraints. Fields used: queryString, numberOfResults, startRow,
   *               and sort fields.
   * @returns {Promise} A Promise for the Passport call.
   */
  searchUsersByQueryString: function(search) {
    //noinspection JSUnresolvedVariable
    const client = this._start()
        .uri('/api/user/search')
        .urlParameter('queryString', search.queryString)
        .urlParameter("numberOfResults", search.numberOfResults)
        .urlParameter("startRow", search.startRow)
        .get();

    //noinspection JSUnresolvedVariable
    if (search.sortFields !== null && typeof search.sortFields !== 'undefined') {
      //noinspection JSUnresolvedVariable
      for (let i = 0; i < search.sortFields.length; i++) {
        //noinspection JSUnresolvedVariable
        client.urlParameter("sortFields[" + i + "].name", search.sortFields[i].name)
            .urlParameter("sortFields[" + i + "].missing", search.sortFields[i].missing)
            .urlParameter("sortFields[" + i + "].order", search.sortFields[i].order);
      }
    }

    return new Promise((resolve, reject) => {
      client.go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Send an email using an email template id. You can optionally provide <code>requestData</code> to access key value
   * pairs in the email template.
   *
   * @param {string} emailTemplateId The id for the template.
   * @param {Object} request         The send email request that contains all of the information used to send the email.
   * @returns {Promise} A Promise for the Passport call.
   */
  sendEmail: function(emailTemplateId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/email/send')
          .urlSegment(emailTemplateId)
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the application with the given id.
   *
   * @param {string} applicationId The id of the application to update.
   * @param {Object} request       The request that contains all of the new application information.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateApplication: function(applicationId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the application role with the given id for the application.
   *
   * @param {string} applicationId The id of the application that the role belongs to.
   * @param {string} roleId        The id of the role to update.
   * @param {Object} request       The request that contains all of the new role information.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateApplicationRole: function(applicationId, roleId, request) {
    this._requireNonNull(applicationId, 'applicationId');
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlSegment('role')
          .urlSegment(roleId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the email template with the given id.
   *
   * @param {string} emailTemplateId The id of the email template to update.
   * @param {Object} request         The request that contains all of the new email template information.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateEmailTemplate: function(emailTemplateId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/email/template')
          .urlSegment(emailTemplateId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the registration for the user with the given id and the application defined in the request.
   *
   * @param {string} userId  The id of the user whose registration is going to be updated.
   * @param {Object} request The request that contains all of the new registration information.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateRegistration: function(userId, request) {
    this._requireNonNull(userId, 'userId');
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/registration')
          .urlSegment(userId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the system configuration.
   *
   * @param {Object} request The request that contains the new SystemConfiguration settings.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateSystemConfiguration: function(request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/system-configuration')
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the user with the given id.
   *
   * @param {string} userId  The id of the user to update.
   * @param {Object} request The request that contains all of the new user information.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateUser: function(userId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the user action with the given id.
   *
   * @param {string} userActionId The id of the user action to update.
   * @param {Object} request      The request that contains all of the new user action information.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateUserAction: function(userActionId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the user action reason with the given id.
   *
   * @param {string} userActionReasonId The id of the user action reason to update.
   * @param {Object} request            The request that contains all of the new user action reason information.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateUserActionReason: function(userActionReasonId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user-action-reason')
          .urlSegment(userActionReasonId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Updates the webhook with the given id.
   *
   * @param {string} webhookId The id of the webhook to update.
   * @param {Object} request   The request that contains all of the new webhook information.
   * @returns {Promise} A Promise for the Passport call.
   */
  updateWebhook: function(webhookId, request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/webhook')
          .urlSegment(webhookId)
          .setJSONBody(request)
          .put()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Validates the provided JWT (encoded JWT string) to ensure the token is valid. A valid access token is properly
   * signed and not expired.
   * <p/>
   * This API may be used to verify the JWT as well as decode the encoded JWT into human readable identity claims.
   *
   * @param {string} encodedJWT The encoded JWT (access token)
   * @return {Promise} A Promise for the Passport call.
   */
  validateAccessToken: function(encodedJWT) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/jwt/validate')
          .authorization('JWT ' + encodedJWT)
          .get()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Verifies the email of a user so they can log in.
   *
   * @param {string} verificationId The uuid sent to a user in an email.
   * @return {Promise} A Promise for the Passport call.
   */
  verifyEmail: function(verificationId) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/user/verify-email')
          .urlSegment(verificationId)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /**
   * Confirms a two factor authentication code.
   *
   * @param {Object} request The two factor request information.
   * @return {Promise} A Promise for the Passport call.
   */
  verifyTwoFactor: function(request) {
    return new Promise((resolve, reject) => {
      this._start()
          .uri('/api/two-factor')
          .setJSONBody(request)
          .post()
          .go(this._responseHandler(resolve, reject));
    });
  },

  /* ===================================================================================================================
   * Private methods
   * ===================================================================================================================*/

  /**
   * Require a parameter to be defined, if null or un-defined this throws an exception.
   * @param {Object} value The value that must be defined.
   * @param {string} name The name of the parameter.
   * @private
   */
  _requireNonNull: function(value, name) {
    if (typeof value === 'undefined' || value === null) {
      throw new Error(name + ' parameter is required.');
    }
  },

  /**
   * Returns a function to handle the promises for each call.
   *
   * @param resolve The promise's resolve function.
   * @param reject The promise's reject function.
   * @returns {Function} The function that will call either the resolve or reject functions based on the ClientResponse.
   * @private
   */
  _responseHandler: function(resolve, reject) {
    return function(response) {
      if (response.wasSuccessful()) {
        resolve(response);
      } else {
        reject(response);
      }
    };
  },

  /**
   * creates a rest client
   *
   * @returns {RESTClient} The RESTClient that will be used to call.
   * @private
   */
  _start: function() {
    return new RESTClient().authorization(this.apiKey).setUrl(this.host);
  }
};

module.exports = PassportClient;